// Written by Folkert van Heusden <mail@vanheusden.com>
// Released under Apache License v2.0

#include <Wire.h>

static bool ungetc_valid = false;
static uint8_t ungetc_byte = 0x00;

void ungetc(const uint8_t c) {
	if (ungetc_valid)
		return; //error(0x09);

	ungetc_byte = c;
	ungetc_valid = true;
}

uint8_t get_byte() {
	if (ungetc_valid) {
		ungetc_valid = false;

		return ungetc_byte;
	}

	static unsigned long int lastChange = 0;
	static bool state = false;
	while(Serial.available() == 0) {
		unsigned long int now = millis();

		if (now - lastChange >= 500) {
			digitalWrite(13, state ? HIGH : LOW);
			lastChange = now;
			state = !state;
		}
	}

	return Serial.read();
}

typedef struct {
	uint8_t midi_command, n_bytes : 2;
} mc_size_t;

const mc_size_t cmds[] = {
	{ 0x80, 2 },
	{ 0x90, 2 },
	{ 0xa0, 2 },
	{ 0xb0, 2 },
	{ 0xc0, 1 },
	{ 0xd0, 1 },
	{ 0xe0, 2 },
	{ 0xf0, 0 },
	{ 0, 0 }
};

bool midi_get(uint8_t *const cmd, uint8_t *const d1, uint8_t *const d2) {
	static uint8_t prev_cmd = 0x00;

	*cmd = get_byte();

	if (*cmd < 0x80) {
		if (!prev_cmd)
			return false;

		ungetc(*cmd);
		*cmd = prev_cmd;
	}

	const uint8_t cmdMasked = (*cmd) & 0xf0;

	*d1 = *d2 = 0x00;

	uint8_t i = 0;
	while(cmds[i].midi_command != cmdMasked && cmds[i].midi_command != 0x00)
		i++;

	if (cmds[i].midi_command == 0x00) {
		//error(0x0a);
		return false;
	}

	if (cmds[i].n_bytes >= 1) {
		*d1 = get_byte();

		if (*d1 & 128) {
			ungetc(*d1);
			*d1 = 0x00;
			//error(0x0b);
			return false;
		}
	}

	if (cmds[i].n_bytes >= 2) {
		*d2 = get_byte();

		if (*d2 & 128) {
			ungetc(*d2);
			*d2 = 0x00;
			//error(0x0c);
			return false;
		}
	}

	if (*cmd < 0xf0)
		prev_cmd = *cmd;

	return true;
}

void setup() {
	Serial.begin(31250);

	pinMode(A4, INPUT_PULLUP);
	pinMode(A5, INPUT_PULLUP);

	Wire.begin();

	for(byte d=1; d<=6; d++) {
		uint8_t msg[4] = { 0, 1, 0, 0 };

		Wire.beginTransmission(d);
		Wire.write(msg, 4);
		Wire.endTransmission();
	}
}

float midiNoteToFreq(const uint8_t note)
{
	return pow(2.0, (float(note) - 69.0) / 12.0) * 440.0;
}

void play_f(uint16_t f, uint16_t v)
{
	for(byte d=1; d<=6; d++) {
		uint16_t this_f = f + d * 9;
		uint8_t msg[4] = { this_f >> 8, this_f & 255, v >> 8, v & 255 };

		Wire.beginTransmission(d);
		Wire.write(msg, 4);
		Wire.endTransmission();
	}
}

uint8_t playing = 255;

void loop() {
	uint8_t cmd, d1, d2;
	if (!midi_get(&cmd, &d1, &d2))
		return;

	if ((cmd & 0x90) == 0x90 && (cmd & 0x0f) != 9) {
		float fr = midiNoteToFreq(d1);

		if (d2 == 0)
			play_f(fr, 0), playing = 255;
		else
			play_f(fr, d2 * 255), playing = d1;
	}
	else if ((cmd & 0x80) == 0x80 && (cmd & 0x0f) != 9) {
		if (d1 == playing) {
			float fr = midiNoteToFreq(d1);

			play_f(fr, 0);

			playing = 255;
		}
	}
}
